#
/*
 *		M U L T I - D E V I C E   Q U E U E I N G   S Y S T E M
 *
 *				  D C O N T R O L . C
 *
 *				Douglas P. Kingston III
 *				  Michael John Muuss
 *
 *				Part of the MDQS Daemon
 *
 *				   Revision History
 *				   ================
 *	13 Sep 82  DPK	Added code for Waveoff's.
 *
 *	14 Sep 82  DPK	Reworked handling of device startup in scandevs.
 *
 *	15 Sep 82  DPK	Removed notify(), in favor of msgopen().
 *
 *	28 Sep 82  DPK	Installed code to support request modifications.
 *			Moved delayed request queueing into queue().
 *
 *	14 Oct 82  DPK	Fixed handling of -2 error return from makereq
 *			in enroll().
 *
 *	5  Mar 1983	David Wilson, BAC.
 *			Don't try to use device if device open failed more
 *			than MAXFAILURES.
 *
 *	10 Mar 1983	David Wilson, BAC.
 *			Fixed bug in update(). When the control file was
 *			being updated, the control file was truncated
 *			by the fopen. Changed to use unbuffered io.
 *
 *	30 Mar 83  DPK	Removed #define MAXFAILURES, and changed references
 *			to MAXFAILURES to the variable Maxfailures.
 *			If Maxfailures is 0, it will try infinitely.
 */

#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include "queue.h"
#include "qreplys.h"
#include <sys/stat.h>

extern	struct queue *quehead;
extern	struct request *delhead;
extern	struct device *devhead;
extern	int delayupdate;
extern	char *Holdprefix;
extern	char *Newprefix;
extern	char *Modprefix;
extern	char *MDQSuser;
extern	int  Maxfailures;
extern	short Waveoff;

struct	queue *getque();
struct	request *gethdr();
FILE	*msgopen();
char	*malloc();
char	*xmalloc();

/*
 *			S C A N D E V S
 *
 *  	Scan through the list of active devices and look
 *	for devices that need to be restarted.  Choose the
 *	next eligible job (if necessary) and call devstart().
 */
scandevs ()
{
	register struct qmap *mp;	/* For stepping through qmaps */
	register struct device *dp;
	register struct request *rp;

	for (dp = devhead; dp != NULL; dp = dp->dv_nextdev) {
		if( dp->dv_status & (D_DISABLED|D_FAILED) )
			continue;		/* Device out of service */

		if( dp->dv_pid != 0 )
			continue;		/* It's already active... */

		else if( dp->dv_actreq != NULL )
			/*
			 *  There is a request assigned to the
			 *  device, it just needs to be restarted.
			 */
			devstart( dp );

		else for (mp=dp->dv_maphd; mp != NULL; mp=mp->m_nextmap) {
			/*
			 *	Its inactive and there is no request
			 *	pending on the device.  Get a new request!
			 */
			if(( rp = mp->m_q->q_readyreq ) == NULL )
				continue;
			if(( dp->dv_status & D_ANYFORM || strncmp(rp->r_forms,
			     dp->dv_forms, FORMSIZ) == SAME)
			     && !(rp->r_status & (SHOLD|SRUN|SKILL))
			     && !delayed(rp)) {
				/* Pull request from the top */
			     	rp->r_status |= SRUN;
				dp->dv_actreq = rp;
				dp->dv_server = mp->m_server;
				dp->dv_actque = mp->m_q;
			     	shuffle( dp, mp );
				devstart( dp );
				break;
			}
			for( rp = rp->r_nextreq; rp != NULL;
				rp = rp->r_nextreq ) {
			    if(( dp->dv_status & D_ANYFORM
			    	|| strncmp(rp->r_forms, dp->dv_forms, FORMSIZ)
				== SAME) && !(rp->r_status & (SHOLD|SRUN|SKILL))
				&& !delayed(rp)) {
				/*
				 *  We found an empty device pointed to by a
				 *  queue with something in it and the forms
				 *  matched, hang request onto device node.
				 */
			     	rp->r_status |= SRUN;
				dp->dv_actreq = rp;
				dp->dv_server = mp->m_server;
				dp->dv_actque = mp->m_q;
				shuffle( dp, mp );
				devstart( dp );
				break;
			    }
			}
		}
	}
}


/*
 *			S H U F F L E
 *
 *	Implements roundrobin scheduling of queues on a device.
 *  Move the last active mapping to the end of the list.
 */
shuffle( dp, mp )
register struct device *dp;
register struct qmap *mp;
{
	register struct qmap *pmp = dp->dv_maphd;

	if (mp->m_nextmap == NULL)
		return;		/* Nothing to shuffle */
	if (mp == pmp) {
		/* Current mapping is at the head */
		dp->dv_maphd = mp->m_nextmap;
		for (; pmp->m_nextmap != NULL; pmp=pmp->m_nextmap);
		mp->m_nextmap = NULL;
		pmp->m_nextmap = mp;
		return;
	}
	for (; pmp->m_nextmap != mp; pmp=pmp->m_nextmap);
	pmp->m_nextmap = mp->m_nextmap;
	for (; pmp->m_nextmap != NULL; pmp=pmp->m_nextmap);
	mp->m_nextmap = NULL;
	pmp->m_nextmap = mp;
	return;
}

/*
 *			E N R O L L
 *
 *	Given a file name, enter it on the proper queue in the
 * right place.  First, the file name is verified for integrity
 * (ie starts with letter 'Q').  Next, the proper device queue
 * is found and the correct priority vicinity is located.
 * Finally, the entry is sorted in by START time.
 *
 *	Note that our current directory must be in the Queue area.
 */

enroll( namep )
register char *namep;
{
	register struct request *rp;
	struct request *makereq();

	/* Verify integrity */
	if( namep[QN_Q] != 'Q')  {
		complain("%s:  not an MDQS file", namep);
		hold (namep);
		return;
	}

	/* Build the request structure */
	rp = makereq( namep );
	if( rp == NULL ) {
		complain ("--- No free requests, skipping... ---\n");
		Waveoff = 1;		/* flag that we are full up */
		return;		
	} else if( rp == (struct request *)-1 )  {
		complain( "makereq failed on %s", namep );
		hold( namep );
		return;
	} else if( rp == (struct request *)-2 ) {
		complain( "Verification error on %s", namep );
		hold( namep );
		return;
	}
	if( queue( rp ) < 0 ) {
		complain( "enroll of %s failed -- no queue", namep );
		zapreq( rp );
		hold( rp->r_name );
		free( rp );
	}
}

/*
 *			Q U E U E
 *
 *	The queue() function takes a request and tries to find
 *	its proper place in the queues.  0 is returned if a place
 *	is found, and -1 if the request was for a nonexistent queue.
 */
queue( rp )
register struct request *rp;
{
	register struct request *r;
	struct request *rprev;
	struct request *rnext;
	register struct queue *qp;
	time_t	timenow;

	/*
	 *  If this is a delayed request put into time sorted list
	 */
	time (&timenow);
	if (timenow < rp->r_tstart) {
		delayupdate = 1;
		if (delhead == NULL || rp->r_tstart < delhead->r_tstart) {
			/*  It goes at the head of the list  */
			rp->r_nextreq = delhead;
			delhead = rp;
			return( 0 );
		}
		rprev = delhead;
		rnext = rprev->r_nextreq;
		for (; rnext != NULL; rprev = rnext, rnext = rnext->r_nextreq)
			if (rp->r_tstart < rnext->r_tstart)
				break;
		rp->r_nextreq = rnext;
		rprev->r_nextreq = rp;
		return( 0 );
	}

	/*
	 *  Must be a "ready" request...
	 *  Find the right logical input queue
	 */
	for (qp = quehead; qp != NULL; qp = qp->q_nextq) {
		if( strcmp( qp->q_name, rp->r_que ) == 0 )  {
			qp->q_status |= Q_UPDATE;

			/* found queue, find position */
			if( (r = qp->q_readyreq) == NULL
			   || r->r_pri > rp->r_pri
			   || (r->r_pri == rp->r_pri &&
			       r->r_tsubmit > rp->r_tsubmit)) {
				/*
				 *  Queue is empty or this is highest priority
				 *
				 *  It goes to the head of the class!
				 */
				rp->r_nextreq = r;
				qp->q_readyreq = rp;
			       	log ("%s queued", rp->r_name);
				return (0);
			}
			while( r->r_nextreq != 0 )  {
			    if( r->r_pri >= rp->r_pri )  {
				/* Here the priority is OK
				 * Two cases to check --
				 * a)  next entry has even lower pri
				 * b)  sort in by start time
				 */
				while( r->r_nextreq->r_pri == rp->r_pri )  {
				    if(r->r_nextreq->r_tsubmit>rp->r_tsubmit) {
					/* It goes here */
					rp->r_nextreq = r->r_nextreq;
					r->r_nextreq = rp;
				       	log ("%s queued", rp->r_name);
					return (0);
				    }
				    r = r->r_nextreq;
				    if( r->r_nextreq == 0 )
					break;
				}

				/* It goes here */
				rp->r_nextreq = r->r_nextreq;
				r->r_nextreq = rp;
			       	log ("%s queued", rp->r_name);
				return (0);
			    }
			    r = r->r_nextreq;
			}
			/* It goes on the end */
			r->r_nextreq = rp;
			rp->r_nextreq = 0;
		       	log ("%s queued", rp->r_name);
			return (0);
		}
	}
	return (-1);
}

/*
 *			M A K E R E Q
 *
 *	This function opens the request control file and
 * reads in & verifies the request structure.
 *
 *	-2 is a verification (security check) error, hold this request
 *	-1 is fatal error return,
 *	 0 indicates that there are no free request structs, try again later,
 *	   otherwise, a request pointer is returned.
 */
struct request *makereq( namep )
register char *namep;
{
	register struct request *rp;
	register FILE *fp;
	struct	stat sb;

	if( (fp = fopen (namep, "r")) == NULL )  {
		complain("Cannot open %s", namep );
		return( (struct request *)-1 );
	}

	/* Get a request structure */
	if(( rp = (struct request *)malloc( sizeof (struct request))) == NULL )  {
		complain("Out of free requests -- skipping %s", namep );
		fclose (fp);
		return(0);
	}

	/* Read in the request structure */
	if( gethdr (fp, rp) == NULL ) {
		fclose (fp);
		complain( "Read error %d on %s", errno, namep );
		zapreq( rp );
		hold (namep);
		free (rp);
		return( (struct request *)-1 );
	}
	fstat( fileno (fp), &sb );
	fclose(fp);

	/* Verify request structure */
	if( strcmp( namep, rp->r_name ) != 0 )  {
		complain( "Verify error on %s/%s", namep, rp->r_name );
		zapreq( rp );
		hold (namep);
		free (rp);
		return( (struct request *)-2 );
	} else if( rp->r_uid != sb.st_uid ) {
		complain( "Security check failure on %s, request held",
			namep );
		zapreq( rp );
		hold( namep );
		free( rp );
		return( (struct request *)-2 );
	}
	return( rp );
}

/*
 *	C H E C K D E L A Y E D
 *
 *	This is called every few seconds by the main loop to move
 *	requests from the delayed request queue to an appropriate
 *	request queue.
 */
checkdelayed ()
{
	register struct request *rp;
	time_t	timenow;
 
	if( delhead == NULL )
		return;

	time( &timenow );
	while( delhead->r_tstart <= timenow ) {
		delayupdate = 1;
		rp = delhead;
		delhead = rp->r_nextreq;
		if( queue( rp ) < 0 ) {
			complain( "queue (%s) failed, %s held",
			   rp->r_name, rp->r_name);
			zapreq( rp );
			hold( rp->r_name );
			free( rp );
		}
		if( delhead == NULL )
			break;
	}
}


/*
 *			H O L D
 *
 *	Take the request control file and put it in the holding tank
 */
hold( rname )
register char	*rname;
{
	register int i;
	char	namebuf[PATHSIZ];

	i = 1;
	do {
		sprintf( namebuf, "%s%s.%d", Holdprefix, rname, i++ );
	} while (link (rname, namebuf) < 0 && errno == EEXIST );
	unlink (rname);
}

/*
 *			R E N E W
 *
 *	Link a request control file back into the new request
 *	directory so that it is recognized again.  Usually the
 *	result of having to throw away a delayed request structure.
 */
renew ( rname )
char	*rname;
{
	char	namebuf[PATHSIZ];

	strcpy (namebuf, Newprefix);
	strcat (namebuf, rname);
	if ( link (rname, namebuf) < 0 ) {
		complain ("Renew link failed (%s: %s)", namebuf,
			  sys_errlist[errno]);
		hold (rname);
	}
}

/*
 *			U P D A T E
 *
 *	This function is given the name of a request which has
 *	been flagged as modified.  It attempts to find the request
 *	and make the appropriate modification.  If it is not found,
 *	or the request cannot be opened, the change is ignored.
 *	(request was probably just finished).
 */
update( rname )
register char	*rname;
{
	char	*modname;
	FILE	*rf;
	struct	request	r;
	register struct	request *rp = NULL;
	register struct	request *prevrp;
	struct	queue *qp;
	time_t	timenow;

debug("update: %s", rname);
	/*
	 *  First open the original control file and use
	 *  that information to locate the request.
	 */
	if(( rf = fopen( rname, "r" )) == NULL )
		return;

	if( gethdr (rf, &r) == (struct request *) NULL )
		return;
	fclose( rf );
	time( &timenow );

	if( r.r_tstart <= timenow ) {
		/*
		 *  There is a chance that the request has been
		 *  been placed on the queue.
		 */
debug("update: check '%.*s'", QUESIZ, r.r_que);
		if(( qp = getque( r.r_que )) == NULL ) {
debug("update: missing queue!");
			goto not_here;
		}
		if(( rp = qp->q_readyreq ) == NULL ) {
debug("update: empty queue.");			
			goto not_here;
		}
		if( strncmp( rp->r_name, rname, QDIRSIZ ) == SAME ) {
			qp->q_readyreq = rp->r_nextreq;
			qp->q_status |= Q_UPDATE;
		} else {
			prevrp = rp;
			for( rp = rp->r_nextreq; rp != NULL;
			     prevrp = rp, rp = rp->r_nextreq )
				if( strncmp(rp->r_name,rname,QDIRSIZ)==SAME ) {
					prevrp->r_nextreq = rp->r_nextreq;
					qp->q_status |= Q_UPDATE;
					break;
				}
		}
	}
not_here:
debug("update: rp is %0o", rp);

	if( rp == NULL && (rp = delhead)) {
		/*
		 *  Request was not in the designated queue.
		 *  The request may not have been queued yet,
		 *  look in the delayed requests list, if non-empty.
		 */
debug("update: check DELAYED");
		if( strncmp( rp->r_name, rname, QDIRSIZ ) == SAME ) {
			delhead = rp->r_nextreq;
			delayupdate++;
		} else {
			prevrp = rp;
			for( rp = rp->r_nextreq; rp != NULL;
		 	    prevrp = rp, rp = rp->r_nextreq )
				if( strncmp(rp->r_name,rname,QDIRSIZ)==SAME ) {
					prevrp->r_nextreq = rp->r_nextreq;
					delayupdate++;
					break;
				}
		}
	}

debug("update: build modname");
	/*
	 *  Now its time to grab the new header since we need
	 *  it to tell if we should go looking in the devices
	 *  if we haven't found the request yet.
	 */
	modname = xmalloc( strlen( Modprefix ) + QDIRSIZ + 2 );
	sprintf( modname, "%s/%s", Modprefix, rname );
	if(( rf = fopen( modname, "r" )) == NULL ) {
		free( modname );
		complain("Can't open modified request file %s", modname);
		return;
	}
	free( modname );

	if( gethdr (rf, &r) == (struct request *) NULL ) {
		fclose( rf );
		complain("Can't read modified request file for %s", rname);
		return;
	}
	fclose( rf );

	/*
	 *  Special handling:
	 *
	 *  For KILL requests, also look at the active devices
	 *  and zap the request there if found.
	 */
	if( rp == NULL && (r.r_status & SKILL)) {
		register struct device *dp;

debug("update: special SKILL processing");
		for( dp = devhead; dp != NULL; dp = dp->dv_nextdev ) {
			if( (rp = dp->dv_actreq) != NULL
			    && strncmp(rname, rp->r_name, QDIRSIZ)==SAME) {
				/*
				 *  Found it!  Mark it for death, and kill
				 *  the current attempt if any.
				 */
				rp->r_status |= SKILL;
debug("update: Found it! (device %s, pid %d)", dp->dv_dname, dp->dv_pid);
				if( dp->dv_pid > 1)
					kill( dp->dv_pid, SIGTERM );
				else {
					char *flushmsg = "Xflushed by qmod -k";

					/*  We caught it idle, expunge it  */
debug("update: Manually expunging...");
					flushmsg[0] = RP_KILLED;
					inform( rp, dp, flushmsg );
					zapreq( rp );
					unlink( rp->r_name );
					free( rp );
					dp->dv_actreq = NULL;
					dp->dv_actque = NULL;
					dp->dv_server = NULL;
					Waveoff = 0;
				}
				return;
			}
		}
	}
	if( rp == NULL )
		return;		/* Sorry Charlie. */
	
debug( "update(): found %s", rname );
	/*
	 *  We have the request, now get lets get to work
	 *  First, the easy case - a kill message
	 */
	if( r.r_status & SKILL ) {
		FILE	*mf;

debug( "update(): cancel %s", rname );
		if(( mf = msgopen( rp )) != NULL ) {
			fprintf( mf, "Your request %s was manually cancelled\n",
				rp->r_name );
			fclose( mf );
		}
		zapreq( rp );
		unlink( rp->r_name );
		free( rp );
		Waveoff = 0;
		return;
	}

	/*
	 *  Update incore copy of request
	 */
	rp->r_tstart = r.r_tstart;
	rp->r_copies = r.r_copies;
	strncpy( rp->r_forms, r.r_forms, FORMSIZ );
	rp->r_options = r.r_options;
	strncpy( rp->r_que, r.r_que, QUESIZ );
	rp->r_pri = r.r_pri;
	rp->r_status = r.r_status;	/** Dangerous Assumptions Made **/

	rupdate(&r);
	if( queue( rp ) < 0 ) {
		complain( "could not requeue modified request %s", rname );
		hold( rname );
	}
}

delayed(rp)
register struct request *rp;
{
	time_t timenow;

	if (rp->r_status & SDELAY) {
		time(&timenow);
		if (timenow < rp->r_tdelay)
			return (1);
		rp->r_status &= (~SDELAY);
		rupdate(rp);
	}
	return (0);
}

rupdate (rp)
register struct request *rp;
{
	register FILE	*rf;

	/*
	 *  Update master copy of information (the control file)
	 */
	if(( rf = fopen( rp->r_name, "r+" )) == NULL ) {
		/*
		 *   Just hope we don't crash...
		 */
		complain( "could not reopen request file %s (%s)",
			   rp->r_name, sys_errlist[errno] );
	} else {
		puthdr (rp, rf);
		fflush (rf);
		if (ferror (rf))
			complain( "error rewriting request file %s (%s)",
				   rp->r_name, sys_errlist[errno] );
		fclose( rf );
	}
}
